#include <stdio.h>
#include <assert.h>
#include <stdint.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "lib/pmc.h"
#include "lib/bench.h"

#if 1
#define xopen(name, flags, mode, fdtable) open(name, flags, mode)
#define xdup(fd, fdtable) dup(fd)
#define xclose(fd, fdtable) close(fd)
#define xfdtable_alloc() 0
#else
#include <sys/syscall.h>
#define __NR_fdtable_alloc                      288
#define __NR_share_open                         289
#define __NR_share_close                        290
#define __NR_share_dup                          291
#define xopen(name, flags, mode, fdtable) syscall(__NR_share_open, name, flags, mode, fdtable)
#define xdup(fd, fdtable) syscall(__NR_share_dup, fd, fdtable)
#define xclose(fd, fdtable) syscall(__NR_share_close, fd, fdtable)
#define xfdtable_alloc() syscall(__NR_fdtable_alloc)
#endif

static char pmc_inited[JOS_NCPU];

enum { fd_ops = 100000 };

typedef struct {
    volatile uint64_t start;
    union {
	struct {
	    volatile uint64_t usec;
	    volatile uint64_t l3miss;
	};
	char pad[JOS_CLINE];
    } cpu[JOS_NCPU] __attribute__((aligned(JOS_CLINE)));
} stats_t;

static stats_t *stats = NULL;

static uint32_t affinity[] = {
    0, 4, 8, 12,
    1, 5, 9, 13,
    2, 6, 10, 14,
    3, 7, 11, 15,
};

static uint32_t fdtable[16];

static void *
fdbang(void *x)
{
    int c = (int)(uintptr_t)x;
    assert(affinity_set(affinity[c]) == 0);
    
    if (!pmc_inited[c]) {
	pmc_init(c);
	pmc_inited[c] = 1;
    }
    
    if (c == 0)
	stats->start = 1;
    else
	while (!stats->start)
	    nop_pause();

    uint64_t s = usec();
    uint64_t sl3 = pmc_l3_miss();
    uint64_t n = 0;

    char fname[32];
    snprintf(fname, sizeof(fname), "foo%u", c);
    int fd0;
    if ((fd0 = xopen(fname, O_WRONLY | O_CREAT | O_TRUNC, S_IRWXU, fdtable[c])) < 0) {
	perror("open");
	exit(-1);
    }
	

    while (1) {
	int fd1;
	assert((fd1 = xdup(fd0, fdtable[c])) >= 0);
	assert(xclose(fd1, fdtable[c]) == 0);

	n++;
	
	if (n == fd_ops) {
	    uint64_t e = usec();
	    uint64_t el3 = pmc_l3_miss();
	    stats->cpu[c].usec = e - s;
	    stats->cpu[c].l3miss = el3 - sl3;
	    break;
	}
    }
    
    return 0;
}

int
main(int ac, char **av)
{
    int max_cores = get_core_count();
    assert(max_cores <= JOS_NCPU);
    pthread_t *th = (pthread_t *)malloc(sizeof(pthread_t) * max_cores);
    affinity_set(0);

    assert(stats = (stats_t *) malloc(sizeof(stats_t)));
    memset(stats, 0, sizeof(*stats));

    for (uint32_t i = 0; i < max_cores; i++) {
        int fd;
	errno_check(fd = xfdtable_alloc());
	fdtable[i] = fd;
    }

    for (int i = 1; i <= max_cores; i++) {
	for (int j = 1; j < i; j++)
	    th[j] = pthread_start(fdbang, (uintptr_t)j);

	usleep(100000);
	fdbang(0);

	for (int j = 1; j < i; j++)
	    pthread_join(th[j], 0);

	uint64_t count = 0;
	uint64_t max = 0;
	uint64_t l3sum = 0;
	for (int j = 0; j < i; j++) {
	    count += stats->cpu[j].usec;
	    l3sum += stats->cpu[j].l3miss;
	    if (stats->cpu[j].usec > max)
		max = stats->cpu[j].usec;
	}

	uint64_t ave = count / i;
	uint64_t per = l3sum / (i * fd_ops);
	uint64_t tp = (fd_ops * i) / (ave / 1000);

	printf(" cores %d, max %ld ms, ave %ld ms, ave op/ms %ld, l3miss/op %ld\n",
	       i, max/1000, ave/1000, tp, per);
	
	memset(stats, 0, sizeof(*stats));
    }
    free(th);
    return 0;
}
