#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <assert.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/resource.h>
#include <sched.h>
#include <pthread.h>
#include <string.h>
#include <signal.h>
#include "bench.h"

//#define USE_KERNEL_SPINLOCK
//#define USE_MCS_LOCK

#if defined(USE_KERNEL_SPINLOCK)

#include "spinlock.h"
typedef raw_spinlock_t testedlock;
#define lock_init(l) __raw_spin_init(l)
#define do_lock(l) __raw_spin_lock(l)
#define do_unlock(l) __raw_spin_unlock(l)

#else

#include "locks.h"
typedef mcslock testedlock;
#define lock_init(l) mcs_init(l)
#define do_lock(l) mcs_lock(l)
#define do_unlock(l) mcs_unlock(l)

#endif

static const int affinity[] = {
	0, 1, 2, 3, 
	4, 5, 6, 7,
	8, 9, 10, 11,
	12, 13, 14, 15,
};

static union {
    testedlock lock;
    char __pad[JOS_CLINE];
} g_lock;

static volatile struct {
    union {
        struct {
	    volatile uint64_t ops;
	    volatile uint64_t running;
	    volatile uint64_t cycles;
        };
	char __pad[JOS_CLINE];
    } cores[JOS_NCPU];
    volatile int state;
} g_state;

enum { CREATED = 1, READY, START, END };
enum { iterations = 100000 };

static void
clean_warm()
{
    do_lock(&g_lock.lock);
    do_unlock(&g_lock.lock);
}

static void*
worker_thread(void *arg)
{
    int cid = (uint64_t)arg;
    if (cid)
        affinity_set(affinity[cid]);
    clean_warm(1);
    if (cid) {
        g_state.state = READY;
        while (g_state.state != START)
	    nop_pause();
    }
    else
	g_state.state = START;
    uint64_t ops = 0;
    uint64_t s = read_tsc();
    while (g_state.state != END) {
        do_lock(&g_lock.lock);
	do_unlock(&g_lock.lock);
        ops ++;
	if (!cid && ops == iterations)
	    g_state.state = END;
    }
    uint64_t e = read_tsc();
    g_state.cores[cid].ops = ops;
    g_state.cores[cid].cycles = e - s;
    g_state.cores[cid].running = 0;
    return 0;
}

static void
ontimer(int sig)
{
    g_state.state = END;
}

int
main(int ac, char **av)
{
    if (ac != 2) {
	printf("Usage: %s <max core number>\n", av[0]);
	exit(-1);
    }
    if (setpriority(PRIO_PROCESS, 0, -20) < 0)
	printf("unable to set process priority\n");
    affinity_set(affinity[0]);
    int ncores = atoi(av[1]);
    memset((void *)&g_state, 0, sizeof(g_state));

    struct itimerval tv;
    tv.it_interval.tv_sec  = 5;
    tv.it_interval.tv_usec = 0;
    tv.it_value = tv.it_interval;
    errno_check(setitimer(ITIMER_REAL, &tv, NULL));
    signal(SIGALRM, ontimer);

    lock_init(&g_lock.lock);

    for (int i = 1; i < ncores; i++) {
	g_state.state = CREATED;
	g_state.cores[i].running = 1;
	pthread_start(worker_thread, i);
	while (g_state.state != READY)
	    nop_pause();
    }
    worker_thread(0);

    uint64_t tot_cpu_cycles = 0;
    uint64_t tot_ops = 0;
    for (int i = 0; i < JOS_NCPU; i++) {
        while (g_state.cores[i].running)
	    nop_pause();
        tot_ops += g_state.cores[i].ops;
	tot_cpu_cycles += g_state.cores[i].cycles;
    }
    double ns = tot_cpu_cycles * (double)1000000000 / (tot_ops * get_cpu_freq());
    printf("%d\t%.0f\n", ncores, ns / ncores);
    return 0;
}
