#include "lib/scopeguard.hh"
#include "lib/error.hh"

extern "C" {
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/mman.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <ifaddrs.h>
#include <sched.h>
#include <stdint.h>

#include "lib/pad.h"
#include "lib/pmc.h"
#include "lib/bench.h"
}

enum { send_bytes = 128 }; // server response size
enum { port_base = 8000 };

enum { do_profile = 1 };
enum { do_oprofile = 0 };
enum { profile_thresh = 10000 };

char send_buf[send_bytes];

static void do_net_test(uint32_t core, uint32_t s_addr, uint16_t port);
static void net_conn(int s);

static struct shared_state {
    PAD(volatile uint64_t) l3miss[JOS_NCPU];
    PAD(volatile uint64_t) conns[JOS_NCPU];
} *shared_state;

#define xpanic(fmt, args...)					\
	do {							\
		fprintf(stderr, "panic: %s:%u " fmt "\n",	\
			__FILE__, __LINE__, ##args);		\
		exit(-1);					\
	} while (0)
#define xsocket socket
#define xbind bind
#define xlisten listen
#define xaccept accept
#define xclose close
#define xwrite write

#define OPROFILE_START "./sh/oprofile.start"
#define OPROFILE_STOP  "./sh/oprofile.stop"

static void
net_conn_helper(void *a)
{
    try {
	int t = (int)(intptr_t)a;
	net_conn((int)t);
    } catch (std::exception &e) {
	printf("net_conn error: %s\n", e.what());
    }
}

static int
create_net_conn(int t)
{
    net_conn_helper((void *)t);
    return 0;
}

int
main(int ac, char **av)
{
    int num_cores = sysconf(_SC_NPROCESSORS_CONF);
    uint32_t ips[64];
    int num_ips = 0;

    void *buf;
    buf = mmap(0, sizeof(*shared_state), 
	       PROT_READ | PROT_WRITE, 
	       MAP_SHARED | MAP_ANONYMOUS, 0, 0);
    assert(buf);
    shared_state = (struct shared_state *)buf;
    memset(shared_state, 0, sizeof(*shared_state));
    
    struct ifaddrs *ifa = NULL, *ifp = NULL;
    if (getifaddrs (&ifp) < 0) {
	perror("getifaddrs");
	return -1;
    }

    for (ifa = ifp; ifa && num_ips < num_cores; ifa = ifa->ifa_next) {
	if(!ifa->ifa_addr)
	    continue;

	if (ifa->ifa_addr->sa_family != AF_INET)
	    continue;
	if ((ifa->ifa_flags & IFF_UP) == 0 ||
	    (ifa->ifa_flags & IFF_LOOPBACK) ||
	    (ifa->ifa_flags & (IFF_BROADCAST | IFF_POINTOPOINT)) == 0) 
	{
	    continue;
	}
	
	uint32_t ip = ((struct sockaddr_in *)ifa->ifa_addr)->sin_addr.s_addr;
	ips[num_ips] = ip;
	num_ips++;
    }

    freeifaddrs (ifp);

    for (int i = 1; i < num_cores; i++) {
	pid_t p = fork();
	if (p < 0)
	    xpanic("fork failed: %s", strerror(errno));
	if (p == 0) {
	    assert(affinity_set(i) == 0);
	    uint32_t port = port_base + i / num_ips;
	    uint32_t ip = ips[i % num_ips];
	    printf("%u starting %u.%u.%u.%u:%u\n",
		   i, 
		   (ip & 0x000000FF), 
		   (ip & 0x0000FF00) >> 8,
		   (ip & 0x00FF0000) >> 16,
		   (ip & 0xFF000000) >> 24,
		   port);
	    do_net_test(i, ips[i % num_ips], htons(port));
	}
    }

    assert(affinity_set(0) == 0);
    uint32_t ip = ips[0];
    uint32_t port = port_base;
    printf("%u starting %u.%u.%u.%u:%u\n",
	   0, 
	   (ip & 0x000000FF), 
	   (ip & 0x0000FF00) >> 8,
	   (ip & 0x00FF0000) >> 16,
	   (ip & 0xFF000000) >> 24,
	   port);
    do_net_test(0, ip, htons(port));
    return 0;
}

static void 
net_conn(int s)
{
    scope_guard<int, int> close_sock(xclose, s);
    int count = send_bytes;
    char *b = send_buf;
    while (count) {
	int r = xwrite(s, b, count);
	if (r < 0)
	    throw basic_exception("write: %s", strerror(errno));
	if (r == 0)
	    throw basic_exception("write: unable to write");
	count -= r;
	b += r;
    }
}

static void __attribute__((noreturn))
do_net_test(uint32_t core, uint32_t s_addr, uint16_t port)
{
    uint64_t l3miss[JOS_NCPU];
    uint64_t conns[JOS_NCPU];

    int s = xsocket(PF_INET, SOCK_STREAM, 0);
    if (s < 0)
	xpanic("unable to create socket: %s", strerror(errno));

    memset(send_buf, 'A' + core, sizeof(send_buf));

    struct sockaddr_in sa;
    bzero(&sa, sizeof(sa));;
    sa.sin_family = AF_INET;
    sa.sin_port = port;
    sa.sin_addr.s_addr = s_addr;

    int reuse = 1;
    if(setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof(reuse)) < 0) 
	xpanic("unable to setsockopt: %s", strerror(errno));

    if (xbind(s, (struct sockaddr*) &sa, sizeof(sa)) < 0) 
	xpanic("unable to bind: %s", strerror(errno));

    if (xlisten(s, 20) < 0) 
	xpanic("unable to listen: %s", strerror(errno));

    if (do_profile)
	pmc_init(core);

    uint64_t long_sum = 0;
    for (uint64_t cnt = 0;; cnt++) {
	int t = xaccept(s, 0, 0);
	if (t < 0) {
	    printf("accept failed %s\n", strerror(errno));
	    continue;
	}

	// start oprofile
	if (do_oprofile && core == 0 && cnt == 0)
	    system(OPROFILE_START);

	int r = create_net_conn(t);
	if (r < 0) {
	    printf("thread_create failed: %s\n", strerror(errno));
	    xclose(t);
	}

	if (do_profile) {
	    shared_state->l3miss[core].v = pmc_l3_miss();

	    shared_state->conns[core].v++;
	    if (core == 1 && ((cnt % 50000) == 0)) {
		uint64_t tconns = 0;
		uint64_t tl3miss = 0;
		for (int i = 0; i < JOS_NCPU; i++) {
		    tconns += shared_state->conns[i].v - conns[i];
		    conns[i] = shared_state->conns[i].v;
		    tl3miss += shared_state->l3miss[i].v - l3miss[i];
		    l3miss[i] = shared_state->l3miss[i].v;
		}

		uint64_t per = tconns ? tl3miss / tconns : 0;
		printf("tconns %ld, tl3miss %ld, tl3miss/tconns %ld\n", 
		       tconns, tl3miss, per);

		long_sum += per;
		if ((cnt % 250000) == 0) {
		    printf(" long miss/conns %ld\n", long_sum / 5);
		    long_sum = 0;
		}
	    }
	}

	if (do_oprofile && cnt == profile_thresh)
	    system(OPROFILE_STOP);
    }
}
